#include "legion.h"
#include "default_mapper.h"
#include "logging_wrapper.h"

using namespace Legion;
using namespace Legion::Mapping;

enum TaskIds
{
  topid,
  readerid
};

static auto io = [](auto msg) { std::cout << msg << std::endl; };

class BuggyMapper : public DefaultMapper
{
public:
  BuggyMapper(Machine, Runtime *, Processor);

  void select_task_options(const MapperContext ctx, const Task &task,
                           TaskOptions &output);

  void map_replicate_task(const MapperContext ctx, const Task &task,
                          const MapTaskInput &input,
                          const MapTaskOutput &def_output,
                          MapReplicateTaskOutput &output);
};

BuggyMapper::BuggyMapper(Machine m, Runtime *rt, Processor p)
    : DefaultMapper(rt->get_mapper_runtime(), m, p)
{
}

void BuggyMapper::select_task_options(const MapperContext ctx, const Task &task,
                                      TaskOptions &output)
{
  DefaultMapper::select_task_options(ctx, task, output);
  output.replicate = true;
}

// Most of this function is copied from the DefaultMapper; see the last part for
// what's new
void BuggyMapper::map_replicate_task(const MapperContext ctx, const Task &task,
                                     const MapTaskInput &input,
                                     const MapTaskOutput &def_output,
                                     MapReplicateTaskOutput &output)
{
  const Processor::Kind target_kind = task.target_proc.kind();
  // Get the variant that we are going to use to map this task
  const VariantInfo chosen = default_find_preferred_variant(
      task, ctx, true /*needs tight bound*/, true /*cache*/, target_kind);

  assert(chosen.is_replicable);

  const std::vector<Processor> &remote_procs =
      remote_procs_by_kind(target_kind);
  // Place on replicate on each node by default
  assert(remote_procs.size() == total_nodes);
  output.task_mappings.resize(total_nodes, def_output);

  // Only check for MPI interop case when dealing with CPUs
  assert(!((target_kind == Processor::LOC_PROC) &&
           runtime->is_MPI_interop_configured(ctx)));

  // Otherwise we can just assign shards based on address space
  // assert(total_nodes > 1);
  for (std::vector<Processor>::const_iterator it = remote_procs.begin();
       it != remote_procs.end(); it++)
  {
    AddressSpace space = it->address_space();
    assert(space < output.task_mappings.size());
    output.task_mappings[space].target_procs.push_back(*it);
  }

  // Indicate that we want to do control replication by filling
  // in the control replication map with our chosen processors
  // Also set our chosen variant
  output.control_replication_map.resize(total_nodes);
  for (unsigned idx = 0; idx < total_nodes; idx++)
  {
    assert(chosen.variant > 0);
    output.task_mappings[idx].chosen_variant = chosen.variant;
    output.control_replication_map[idx] =
        output.task_mappings[idx].target_procs[0];

    // This is the only thing that's different from the DefaultMapper
    if (!strcmp(task.get_task_name(), "reader"))
    {
      for (int i = 0; i < task.regions.size(); i++)
      {
        const TaskLayoutConstraintSet &layout_constraints =
            runtime->find_task_layout_constraints(ctx, task.task_id,
                                                  chosen.variant);

        auto memory_constraint =
            find_memory_constraint(ctx, task, chosen.variant, i);

        auto target_memory = default_policy_select_target_memory(
            ctx, output.task_mappings[idx].target_procs[0], task.regions[i],
            memory_constraint);

        size_t footprint;
        std::set<FieldID> copy = task.regions[i].privilege_fields;

        assert(default_create_custom_instances(
            ctx, output.task_mappings[idx].target_procs[0], target_memory,
            task.regions[i], i, copy, layout_constraints, true,
            output.task_mappings[idx].chosen_instances[i], &footprint));
      }
    }
  }
}

void mapper_registration(Machine machine, Runtime *rt,
                         const std::set<Processor> &local_procs)
{
  for (std::set<Processor>::const_iterator it = local_procs.begin();
       it != local_procs.end(); it++)
  {
    rt->replace_default_mapper(
        new LoggingWrapper(new BuggyMapper(machine, rt, *it)), *it);
  }
}

void reader(const Task *task, const std::vector<PhysicalRegion> &regions,
            Context ctx, Runtime *rt)
{
  return;
}

void top(const Task *task, const std::vector<PhysicalRegion> &regions,
         Context ctx, Runtime *rt)
{
  Rect<1> domain(0, 1);
  auto ispace = rt->create_index_space(ctx, domain);
  auto fspace = rt->create_field_space(ctx);
  {
    auto fal = rt->create_field_allocator(ctx, fspace);
    fal.allocate_field(sizeof(int), 0);
  }
  auto lr = rt->create_logical_region(ctx, ispace, fspace);

  rt->fill_field<int>(ctx, lr, lr, 0, 0);

  TaskLauncher rl(readerid, TaskArgument());
  rl.add_region_requirement(RegionRequirement(lr, READ_ONLY, EXCLUSIVE, lr));
  rl.add_field(0, 0);
  rt->execute_task(ctx, rl);
}

int main(int argc, char **argv)
{
  Runtime::set_top_level_task_id(topid);

  {
    TaskVariantRegistrar registrar(topid, "top");
    registrar.add_constraint(ProcessorConstraint(Processor::LOC_PROC));
    registrar.set_replicable(true);
    Runtime::preregister_task_variant<top>(registrar, "top");
  }
  {
    TaskVariantRegistrar registrar(readerid, "reader");
    registrar.add_constraint(ProcessorConstraint(Processor::LOC_PROC));
    registrar.set_replicable(true);
    Runtime::preregister_task_variant<reader>(registrar, "reader");
  }

  Runtime::add_registration_callback(mapper_registration);
  return Runtime::start(argc, argv);
}
